/*
** $Id: dcodemh.c,v 1.1 2002/04/12 06:29:28 sysuh Exp sysuh $
*/

/*
 * Library functions
 *
 * char *dcodeMH( char *ibuf )
 *    decode the B64 or Q encoded MIME string.
 * unsigned char *hsym2asc( unsigned char *str )
 *    convert hangul symbols in page 21, 22, 23 to ASCII symbols.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>

#define	KS5601_1(c)	((c) >= 0xa1 && (c) <= 0xfd)
#define	KS5601_2(c)	((c) >= 0xa1 && (c) <= 0xfe)

/*	RFC-1342 header encoding start/end sequence	*/
#define PREFIX		"=?" 
#define	POSTFIX	 	"?="
#define SUFFIXQP	"?Q?" 
#define SUFFIXB64	"?B?" 
#define OLDPREFIX	"=?B?EUC-KR?" 
#define HEADER_CTE	"Content-Transfer-Encoding: "
#define HEADER_CT	"Content-Type: "
#define	KOR_CHARSET	"EUC-KR"	/* KSC-5601	*/

/*	ISO-2022 encoding designator escape sequence	*/
#define	INTRO_ISO	"\033$)C"
#define	SHIFTOUT	'\016'		/* ASCII SO	*/
#define	SHIFTIN		'\017'		/* ASCII SI	*/
#define	DEL		'\177'		/* ASCII DEL	*/
#define OFFSET	( unsigned char ) 0200	/* or ISO encoding offset */

#define	LSIZ		4096
#define	TRUE		1
#define	FALSE		0

#define PAD			'='
#define	CHAR64(c)	strchr(b64seq, (c)) ? strchr(b64seq, (c)) - b64seq : -1;
static char *b64seq =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/* Decode base64 encoded string */
static int decodeB64Str( ibuf, obuf, nc )
char ibuf[], obuf[];
int nc;
{
    char *iptr = ibuf;
    char *optr = obuf;
    int c1, c2, c3, c4;

    assert( ibuf !=NULL && obuf != NULL && nc >= 0 );

    while( (c1 = *iptr++) && iptr <= &ibuf[nc] ) {
	if( isascii(c1) && isspace(c1))
	    continue;

	do {
	    c2 = *iptr++;
	} while( iptr <= &ibuf[nc] && isascii(c2) && isspace(c2));
	if( iptr > &ibuf[nc] || c2 == '\0' )
	    break;

	do {
	    c3 = *iptr++;
	} while( iptr <= &ibuf[nc] && isascii(c3) && isspace(c3));
	if( iptr > &ibuf[nc] || c3 == '\0' )
	    break;

	do {
	    c4 = *iptr++;
	} while( iptr <= &ibuf[nc] && isascii(c4) && isspace(c4));
	if( iptr > &ibuf[nc] || c4 == '\0' )
	    break;

	if( c1 == PAD || c2 == PAD )
	    continue;

	c1 = CHAR64(c1);
	c2 = CHAR64(c2);
	if( c1 < 0 || c2 < 0 )
	    return -1;
	*optr++ = (c1 << 2) | ((c2 & 0x30) >> 4);

	if( c3 == PAD )
	    continue;
	c3 = CHAR64(c3);
	if( c3 < 0 )
	    return -1;
	*optr++ = ((c2 & 0x0f) << 4) | ((c3 & 0x3c) >> 2);

	if( c4 == PAD )
	    continue;
	c4 = CHAR64(c4);
	if( c4 < 0 )
	    return -1;
	*optr++ = ((c3 & 0x03) << 6) | c4;
    }
    *optr = '\0';
    return 0;
}

/* Convert two hexadecimal digit char to integer  	*/
/*   ex)  h2toi('4', 'E') == 0x4e == 'N' 		*/
/* If any char is not hexadecimal digit, return -1	*/
static int h2toi( ch1, ch2 )
char ch1, ch2 ;
{
    unsigned d1, d2 ;

    if ((ch1) >= '0' && (ch1) <= '9' ) d1 = ( ch1 - '0' );
    else if ((ch1) >= 'A' && (ch1) <= 'F' ) d1 = ( 10 + ch1 - 'A' );
    else if ((ch1) >= 'a' && (ch1) <= 'f' )  d1 = ( 10 + ch1 - 'a' );
    else return(-1);

    if ((ch2) >= '0' && (ch2) <= '9' ) d2 = ( ch2 - '0' );
    else if ((ch2) >= 'A' && (ch2) <= 'F' ) d2 = ( 10 + ch2 - 'A' );
    else if ((ch2) >= 'a' && (ch2) <= 'f' )  d2 = ( 10 + ch2 - 'a' );
    else return(-1);
 
    return ( (int) (((d1<<4) + d2 ) & 0xff)) ;
}

static int decodeQPStr(ibuf, obuf, limit)
char ibuf[], obuf[];
int limit;
{
    char *iptr = ibuf, *optr = obuf;
    int ctmp;
    assert( iptr != NULL && obuf != NULL && limit >= 0 );

    while( *iptr && ( iptr - ibuf < limit ) ) {
	if ( *iptr == '=' && ( ctmp = h2toi( iptr[1], iptr[2])) >= 0 ){
	    *optr++ = (char) ctmp;
	    iptr += 3;
	}
	else if ( *iptr == '_' ) {  /* 	translate    */
	    *optr++ = ' '; iptr++;
	}
	else
	    *optr++ = *iptr++;
    }
    *optr = '\0';
    return 0;
}

/*
** Decode a MIME header encoded by RFC1342 MIME "Q" or "B" header encoding rule.
** Returns the address of a constant character string. Don't free it.
*/
char *dcodeMH( ibuf )
char *ibuf;
{
    static char *obuf;
    static int len_obuf;
    char *iptr, *optr;
    char *preptr, *sufptr, *txtptr, *postptr ;
    int n;
    int found_encoding = 0;

    assert( ibuf != NULL );

    /* make sure we have big enough output buffer */
    n = strlen( ibuf ) + 1;
    if( len_obuf < n ) {
	if( obuf )
	    free( obuf );
	obuf = malloc( n );
	assert( obuf );
	len_obuf = n;
    }
    iptr = ibuf;
    optr = obuf;

    while( *iptr ) {
	if ( ( preptr = strstr( iptr, PREFIX )) == NULL
	    || ( sufptr = strchr( preptr+strlen(PREFIX), '?' )) == NULL 
	    /*   misssing POSTFIX, do not decode   */
	    || ( postptr = strstr( sufptr+strlen(SUFFIXQP), POSTFIX ))==NULL
	    /*   Don't decode UTF-7 or UTF-8  */
	    || strncasecmp(preptr+2, "utf", 3) == 0
	) {
	    /* no more encoded string */
	    strcpy( optr, iptr );
	    return obuf;
	}

	txtptr = sufptr+strlen(SUFFIXQP) ; 
	/* (header)  =?EUC-KR?Q?=89=AB=CD=EF?=       */
	/* (ptr's)   ^pre    ^sf^txt        ^post    */ 

	/* rfc2047: ignore WSP chars between encoded chars */
	n = preptr - iptr;
	if( !found_encoding || strspn( iptr, " \t\r\n" ) != n )
	{
	    strncpy( optr, iptr, n ) ;
	    optr += n ;    /* watch out order */
	}
	iptr = preptr ;

	if( strncasecmp( sufptr, SUFFIXQP, strlen(SUFFIXQP) ) == 0 ) {
	    if( decodeQPStr( txtptr , optr, postptr-txtptr ) < 0 )
		return ibuf;
	    found_encoding = 1;
	}
	else if( strncasecmp( sufptr, SUFFIXB64, strlen(SUFFIXB64) ) == 0 ) {
	    if( decodeB64Str( txtptr , optr, postptr-txtptr ) < 0 )
		return ibuf;	/* error */
	    found_encoding = 1;
	}
	/* For compatibility with old (before Dec. 94) elm2.3h or hcode2.0  */
	else if( strncasecmp( preptr, OLDPREFIX, strlen(OLDPREFIX)) == 0 ) {
	    txtptr = preptr+strlen(OLDPREFIX) ; 
	    if( decodeB64Str( txtptr , optr, postptr-txtptr ) < 0 )
		return ibuf;
	}
	else {   /* Unknown encoding, do not decode */
	    return ibuf;
	}

	optr += strlen(optr) ;
	iptr = postptr + strlen(POSTFIX);
    }
    *optr = '\0';
    return obuf;
}

static char cp21[] =
{
	0, 0, 0, 0,		0, 0, 0, 0,
			0, 0, 0, 0,		0, 0, 0, 0,
	0, 0, 0, 0,		0, 0, 0, 0,
			0, 0, 0, 0,		0, 0, 0, 0,
	0, ' ', '#', '#',  '.', '#', '#', '#',
			'"', '-', '_', '|',		'\\', '~', '`', '\'',
	'"', '"', '[', ']',    '<', '>', '<', '>',
			'[', ']', '[', ']',		'[', ']', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'"', '#', '#', '#',		'#', '#', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'#', '#', '#', '#',		'#', '#', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'#', '#', '#', '#',		'#', '#', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'#', '#', '#', '#',		'#', '#', '#', 0
};

static char cp22[] =
{
	0, 0, 0, 0,		0, 0, 0, 0,
			0, 0, 0, 0,		0, 0, 0, 0,
	0, 0, 0, 0,		0, 0, 0, 0,
			0, 0, 0, 0,		0, 0, 0, 0,
	0, '#', '#', '#',		'#', '\'', '~', '#',
			'#', '"', '#', '#',		'>', '<', '!', '#',
	':', '#', '#', '#',		'#', '#', '%', '#',
			'#', '#', '#', '#',		'#', '#', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'#', '#', '#', '#',		'#', '#', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'#', '#', '#', '#',		'#', '#', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'#', '#', '#', '#',		'#', '#', '#', '#',
	'#', '#', '#', '#',		'#', '#', '#', '#',
			'#', '#', '#', '#',		'#', '#', '#', 0
};

static char cp23[] =
{
	0, 0, 0, 0,		0, 0, 0, 0,
			0, 0, 0, 0,		0, 0, 0, 0,
	0, 0, 0, 0,		0, 0, 0, 0,
			0, 0, 0, 0,		0, 0, 0, 0,
	0, '!', '"', '#',		'$', '%', '&', '\'',
			'[', ']', '*', '+',		',', '-', '.', '/',
	'0', '1', '2', '3',		'4', '5', '6', '7',
			'8', '9', ':', ';',		'<', '=', '>', '?',
	'@', 'A', 'B', 'C',		'D', 'E', 'F', 'G',
			'H', 'I', 'J', 'K',		'L', 'M', 'N', 'O',
	'P', 'Q', 'R', 'S',		'T', 'U', 'V', 'W',
			'X', 'Y', 'Z', '[',		'$', ']', '^', '_',
	'`', 'a', 'b', 'c',		'd', 'e', 'f', 'g',
			'h', 'i', 'j', 'k',		'l', 'm', 'n', 'o',
	'p', 'q', 'r', 's',		't', 'u', 'v', 'w',
			'x', 'y', 'z', '{',		'|', '}', '~', 0
};

/*
** Convert Hangul(KS5601) symbols chars to ASCII.
** Replace it to blank if there's no equivalent ASCII symbol.
** Return a static string.  Do not deallocate it.
*/
unsigned char *hsym2asc( str )
unsigned char *str;
{
    static unsigned char *buf;
    static int len;
    unsigned char *s, *h;
    int hangul = 0;
    int c, l;

    /* compute the string lengh. */
    for( s=str; (c=*s) != 0; s++ )
	if( c >= 128 )
	    hangul = 1;
    /* if there's no Hangul character, return the original string */
    if( !hangul )
	return str;

    /* allocate output buffer */
    l = s - str + 1;
    if( l > len ) {
	if( buf )
	    free( buf );
	buf = (unsigned char *)malloc( l );
	assert( buf );
	len = l;
    }

    /* check and copy the character one by one */
    for( s=str,h=buf; (c=*s) != 0; s++ ) {
	unsigned char d = *(s+1);
	if( KS5601_1(c) && KS5601_2(d)) {
	    if( c == 0xa1 )
		*h++ = cp21[d-128];
	    else if( c == 0xa2 )
		*h++ = cp22[d-128];
	    else if( c == 0xa3 )
		*h++ = cp23[d-128];
	    else {
		*h++ = c;
		*h++ = d;
	    }
	    s++;
	}
	else
	    *h++ = c;
    }
    *h = 0;
    return buf;
}

#ifdef	TEST
#define	BUFSIZE	8192

static char *pname;
static void usage( void )
{
    fprintf( stderr, "usage: %s [infile]\n", pname );
    exit(1);
}

int main( int argc, char **argv )
{
    char	*infile = NULL;
    char	buf[BUFSIZE];
    int		i, n;
    char	*p;

    pname = argv[0];
    for (i=1; i<argc; i++)
	if( *argv[i] != '-' && infile == NULL )
	    infile = argv[i];
	else
	    usage();
    if( infile && !freopen( infile, "r", stdin ) ) {
	fprintf( stderr, "can't open %s to read\n", infile );
	exit(1);
    }

    if (infile && !freopen(infile, "r", stdin)) {
	fprintf(stderr, "can't open input %s\n", infile);
	exit(1);
    }

    for( i=0; i<BUFSIZ-10; ) {
	if( fgets( &buf[i], BUFSIZ-i, stdin) == NULL )
	    break;
	n = strlen(&buf[i]);
	if( n == 1 && buf[i] == '\n' )
	    break;
	i += n;
    }
    /* delete trailing newlines */
    i = strlen(buf);
    while( i > 0 && buf[i-1] == '\n' )
	buf[--i] = '\0';
    p = dcodeMH( buf );
    printf( "%s\n", hsym2asc(p) );
    return 0;
}
#endif	/* TEST */
